@@ -9,16 +9,16 @@
 
 
 
-//if first item added to Include/Exclude is an include
-//then the program will exclude by default
-void AddIncludeExclude(int Type, const char *Item)
+void AddIncludeExclude(HashratCtx *Ctx, int Type, const char *Item)
 {
 ListNode *Node;
 
+//if we get given an include with no previous excludes, then 
+//set CTX_EXCLUDE as the default
 if (! IncludeExclude)
 {
   IncludeExclude=ListCreate();
-  if (Type==FLAG_INCLUDE) Flags |= FLAG_EXCLUDE;
+  if (Type==CTX_INCLUDE) Ctx->Flags |= CTX_EXCLUDE;
 }
 
 Node=ListAddItem(IncludeExclude, CopyStr(NULL, Item));
@@ -65,7 +65,7 @@ DestroyString(Tempstr);
 
 int CommandLineHandleArg(int argc, char *argv[], int pos, int ParseFlags, int SetFlags, char *VarName, char *VarValue, ListNode *Vars)
 {
-	Flags |= (SetFlags & ~(FLAG_INCLUDE | FLAG_EXCLUDE));
+	Flags |= SetFlags;
 	
 	if (ParseFlags & CMDLINE_ARG_NAMEVALUE)
 	{
@@ -78,9 +78,7 @@ int CommandLineHandleArg(int argc, char *argv[], int pos, int ParseFlags, int Se
 	{
 		strcpy(argv[pos],"");
 		pos++;
-		if (SetFlags & FLAG_INCLUDE) AddIncludeExclude(FLAG_INCLUDE, argv[pos]);
-		else if (SetFlags & FLAG_EXCLUDE) AddIncludeExclude(FLAG_EXCLUDE, argv[pos]);
-		else SetVar(Vars,VarName,argv[pos]);
+		SetVar(Vars,VarName,argv[pos]);
 	}
 	}
 	else if (StrLen(VarName)) SetVar(Vars,VarName,VarValue);
@@ -92,11 +90,25 @@ return(ParseFlags);
 
 
 
-void CommandLineSetCtx(int argc, char *argv[], int i, HashratCtx *Ctx, int Flag, int Encoding)
+void CommandLineSetCtx(int argc, char *argv[], int pos, HashratCtx *Ctx, int Flag, int Encoding)
 {
 if (Encoding > 0) Ctx->Encoding=Encoding;
 Ctx->Flags |= Flag;
-strcpy(argv[i],"");
+strcpy(argv[pos],"");
+
+if (Flag == CTX_INCLUDE) 
+{
+	pos++;
+	AddIncludeExclude(Ctx,CTX_INCLUDE, argv[pos]);
+	strcpy(argv[pos],"");
+}
+else if (Flag == CTX_EXCLUDE) 
+{
+	pos++;
+	AddIncludeExclude(Ctx,CTX_EXCLUDE, argv[pos]);
+	strcpy(argv[pos],"");
+}
+
 }
 
 
@@ -126,6 +138,7 @@ strcpy(argv[i],"");
 DestroyString(Token);
 }
 
+
 //this is the main parsing function that goes through the command-line args
 HashratCtx *CommandLineParseArgs(int argc,char *argv[])
 {
@@ -206,14 +219,26 @@ else if (
 	Ctx->Action=ACT_PRINTUSAGE;
 	return(Ctx);
 }
-else if (strcmp(argv[i],"-c")==0)
+else if (strcmp(argv[i],"-C")==0)
 {
 	Ctx->Action = ACT_CHECK;
+	CommandLineSetCtx(argc, argv, i, Ctx, CTX_RECURSE,0);
+}
+else if (strcmp(argv[i],"-Cf")==0)
+{
+	Ctx->Action = ACT_CHECK;
+	CommandLineSetCtx(argc, argv, i, Ctx, CTX_RECURSE,0);
+	ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_OUTPUT_FAILS, "", "",Ctx->Vars);
+	strcpy(argv[i],"");
+}
+else if (strcmp(argv[i],"-c")==0)
+{
+	Ctx->Action = ACT_CHECK_LIST;
 	strcpy(argv[i],"");
 }
 else if (strcmp(argv[i],"-cf")==0)
 {
-	Ctx->Action = ACT_CHECK;
+	Ctx->Action = ACT_CHECK_LIST;
 	ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_OUTPUT_FAILS, "", "",Ctx->Vars);
 }
 else if (strcmp(argv[i],"-s")==0)
@@ -261,6 +286,12 @@ else if (strcmp(argv[i],"-cB")==0)
 	Ctx->Action = ACT_CHECKBACKUP;
 	strcpy(argv[i],"");
 }
+else if ((strcmp(argv[i],"-dir")==0) || (strcmp(argv[i],"-dirmode")==0))
+{
+	Ctx->Action = ACT_HASHDIR;
+	Ctx->Flags |= CTX_RECURSE;
+	strcpy(argv[i],"");
+}
 else if ((strcmp(argv[i],"-hook")==0) || (strcmp(argv[i],"-h")==0))
 {
 	strcpy(argv[i],"");
@@ -297,6 +328,12 @@ else if (strcmp(argv[i],"-jh256")==0) ParseFlags |= CommandLineHandleArg(argc, a
 else if (strcmp(argv[i],"-jh384")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, 0, "HashType", "jh-384",Ctx->Vars);
 else if (strcmp(argv[i],"-jh512")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, 0, "HashType", "jh-512",Ctx->Vars);
 else if (strcmp(argv[i],"-jh")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, 0, "HashType", "jh-512",Ctx->Vars);
+else if (strcmp(argv[i],"-type")==0) 
+{
+	strcpy(argv[i],"");
+	i++;
+	ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, 0, "HashType", argv[i],Ctx->Vars);
+}
 //else if (strcmp(argv[i],"-crc32")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, 0, "HashType", "crc32",Ctx->Vars);
 else if (strcmp(argv[i],"-8")==0)  CommandLineSetCtx(argc, argv, i, Ctx,  0, ENCODE_OCTAL);
 else if (strcmp(argv[i],"-10")==0) CommandLineSetCtx(argc, argv, i, Ctx,  0, ENCODE_DECIMAL);
@@ -315,21 +352,20 @@ else if (strcmp(argv[i],"-z85")==0) CommandLineSetCtx(argc, argv, i, Ctx,  0, EN
 else if (strcmp(argv[i],"-d")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_DEREFERENCE,0);
 else if (strcmp(argv[i],"-X")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_EXES,0);
 else if (strcmp(argv[i],"-exe")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_EXES,0);
+else if (strcmp(argv[i],"-r")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_RECURSE,0);
+else if (strcmp(argv[i],"-fs")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_ONE_FS,0);
 else if (strcmp(argv[i],"-n")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE, 0, "Output:Length", "",Ctx->Vars);
 else if (strcmp(argv[i],"-hmac")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE, FLAG_HMAC, "EncryptionKey", "",Ctx->Vars);
 else if (strcmp(argv[i],"-idfile")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE, 0,  "SshIdFile", "",Ctx->Vars);
-else if (strcmp(argv[i],"-r")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_RECURSE, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-f")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_FROM_LISTFILE, 0, "", "",Ctx->Vars);
-else if (strcmp(argv[i],"-i")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE, FLAG_INCLUDE , "", "",Ctx->Vars);
-else if (strcmp(argv[i],"-x")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE, FLAG_EXCLUDE , "", "",Ctx->Vars);
-else if (strcmp(argv[i],"-dirmode")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_DIRMODE | FLAG_RECURSE, "", "",Ctx->Vars);
-else if (strcmp(argv[i],"-devmode")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_DIRMODE | FLAG_DEVMODE, "", "",Ctx->Vars);
+else if (strcmp(argv[i],"-i")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_INCLUDE,0);
+else if (strcmp(argv[i],"-x")==0) CommandLineSetCtx(argc, argv, i, Ctx, CTX_EXCLUDE,0);
+else if (strcmp(argv[i],"-devmode")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_DEVMODE, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-lines")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_LINEMODE, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-rawlines")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_RAW|FLAG_LINEMODE, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-hide-input")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_HIDE_INPUT, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-star-input")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_STAR_INPUT, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-rl")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_RAW|FLAG_LINEMODE, "", "",Ctx->Vars);
-else if (strcmp(argv[i],"-fs")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_ONE_FS, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-xattr")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_XATTR, 0, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-txattr")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_TXATTR, 0, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-cache")==0) CommandLineSetCtx(argc, argv, i, Ctx,   CTX_XATTR_CACHE,0);
@@ -339,6 +375,7 @@ else if (strcmp(argv[i],"-S")==0) ParseFlags |= CommandLineHandleArg(argc, argv,
 else if (strcmp(argv[i],"-net")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_NET, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-memcached")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE|CMDLINE_MEMCACHED, 0, "Memcached:Server", "",Ctx->Vars);
 else if (strcmp(argv[i],"-mcd")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, CMDLINE_ARG_NAMEVALUE| CMDLINE_MEMCACHED, 0, "Memcached:Server", "",Ctx->Vars);
+else if (strcmp(argv[i],"-xsel")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_XSELECT, "", "",Ctx->Vars);
 else if (strcmp(argv[i],"-v")==0) ParseFlags |= CommandLineHandleArg(argc, argv, i, 0, FLAG_VERBOSE, "", "",Ctx->Vars);
 else if (
 					(strcmp(argv[i],"-t")==0) ||
@@ -401,6 +438,7 @@ else Ctx->HashType=CopyStr(Ctx->HashType,GetVar(Ctx->Vars,"HashType"));
 switch (Ctx->Action)
 {
 case ACT_CHECK:
+case ACT_CHECK_LIST:
 	if (ParseFlags & CMDLINE_XATTR) Ctx->Action=ACT_CHECK_XATTR;
 	if (ParseFlags & CMDLINE_MEMCACHED) Ctx->Action=ACT_CHECK_MEMCACHED;
 break;
@@ -490,7 +528,9 @@ printf("  %-15s %s\n","-i <pattern>", "Only hash items matching <pattern>");
 printf("  %-15s %s\n","-x <pattern>", "Exclude items matching <pattern>");
 printf("  %-15s %s\n","-n <length>", "Truncate hashes to <length> bytes");
 printf("  %-15s %s\n","-c", "CHECK hashes against list from file (or stdin)");
-printf("  %-15s %s\n","-cf", "CHECK hashes but only show failures");
+printf("  %-15s %s\n","-cf", "CHECK hashes against list but only show failures");
+printf("  %-15s %s\n","-C <dir>", "Recursively CHECK directory against list of files on stdin ");
+printf("  %-15s %s\n","-Cf <dir>", "Recursively CHECK directory against list but only show failures");
 printf("  %-15s %s\n","-m", "MATCH files from a list read from stdin.");
 printf("  %-15s %s\n","-lm", "Read hashes from stdin, upload them to a memcached server (requires the -memcached option).");
 printf("  %-15s %s\n","-X", "In CHECK or MATCH mode only examine executable files.");
@@ -505,6 +545,7 @@ printf("  %-15s %s\n","-strict", "Strict mode: when checking, check file mtime,
 printf("  %-15s %s\n","-S", "Strict mode: when checking, check file mtime, owner, group, and inode as well as it's hash");
 printf("  %-15s %s\n","-d","dereference (follow) symlinks"); 
 printf("  %-15s %s\n","-fs", "Stay on one file system");
+printf("  %-15s %s\n","-dir", "DirMode: Read all files in directory and create one hash for them!");
 printf("  %-15s %s\n","-dirmode", "DirMode: Read all files in directory and create one hash for them!");
 printf("  %-15s %s\n","-devmode", "DevMode: read from a file EVEN OF IT'S A DEVNODE");
 printf("  %-15s %s\n","-lines", "Read lines from stdin and hash each line independantly.");
@@ -515,13 +556,13 @@ printf("  %-15s %s\n","-net", "Treat 'file' arguments as either ssh or http URLs
 printf("  %-15s %s\n","", "URLs are in the format ssh://[username]:[password]@[host]:[port] or http://[username]:[password]@[host]:[port]..");
 printf("  %-15s %s\n","-idfile <path>", "Path to an ssh private key file to use to authenticate INSTEAD OF A PASSWORD when pulling files via ssh.");
 printf("  %-15s %s\n","-xattr", "Use eXtended file ATTRibutes. In hash mode, store hashes in the file attributes, in check mode compare against hashes stored in file attributes.");
-printf("  %-15s %s\n","-txattr", "Use TRUSTED eXtended file ATTRibutes. In hash mode, store hashes in 'trusted' file attributes. 'trusted' attributes can only be read and written by root.");
+printf("  %-15s %s\n","-txattr", "Use TRUSTED eXtended file ATTRibutes. In hash mode, store hashes in 'trusted' file attributes. 'trusted' attributes can only be read and written by root. Under freebsd this menas SYSTEM attributes.");
 printf("  %-15s %s\n","-attrs", "comma-separated list of filesystem attribute names to be set to the value of the hash.");
 printf("  %-15s %s\n","-cache", "Use hashes stored in 'user' xattr if they're younger than the mtime of the file. This speeds up outputting hashes.");
 printf("  %-15s %s\n","-u <types>", "Update. In checking mode, update hashes for the files as you go. <types> is a comma-separated list of things to update, which can be 'xattr' 'memcached' or a file name. This will update these targets with the hash that was found at the time of checking.");
 printf("  %-15s %s\n","-hide-input", "When reading data from stdin in linemode, set the terminal to not echo characters, thus hiding typed input.");
 printf("  %-15s %s\n","-star-input", "When reading data from stdin in linemode replace characters with stars.");
-
+printf("  %-15s %s\n","-xsel", "Update X11 clipboard and primary selections to the current hash. This works using Xterm command sequences. The xterm resource 'allowWindowOps' must be set to 'true' for this to work.");
 
 printf("\n\nHashrat can also detect if it's being run under any of the following names (e.g., via symlinks)\n\n");
 printf("  %-15s %s\n","md5sum","run with '-trad -md5'");
