@@ -2,8 +2,9 @@
 #include "DataProcessing.h"
 #include "pty.h"
 #include "expect.h"
-#include <sys/mman.h>
 #include <sys/file.h>
+#include "securemem.h"
+#include <sys/mman.h>
 
 #ifdef HAVE_LIBSSL
 #include <openssl/crypto.h>
@@ -84,19 +85,36 @@ void STREAMSetFlags(STREAM *S, int Set, int UnSet)
 {
 int val;
 
-        S->Flags &= ~UnSet;
-        S->Flags |= Set;
+  S->Flags &= ~UnSet;
+  S->Flags |= Set;
 
-        //Handling nonblock flag involves setting nonblock on or off
-  fcntl(S->in_fd,F_GETFL,&val);
+  //yes, flags are RETURNED, argument ignored
+  val=fcntl(S->in_fd,F_GETFL,NULL);
+
+  //Handling nonblock flag involves setting nonblock on or off
+  //sadly fcntl does not honor O_CLOEXEC flag, so we have to set
+  //that using F_SETFD below
   if (S->Flags & SF_NONBLOCK) val |= O_NONBLOCK;
   else val &= (~O_NONBLOCK);
+
+  //these two sets of flags should be the same
   fcntl(S->in_fd, F_SETFL, val);
+  fcntl(S->out_fd, F_SETFL, val);
+
+
+  //F_GETFD, NOT F_GETFL as above
+  val=fcntl(S->in_fd,F_GETFD, NULL);
+  //Handling close-on-exec flag
+  if (S->Flags & SF_EXEC_INHERIT) val &= (~FD_CLOEXEC);
+  else val |= FD_CLOEXEC;
+
+  //NOTE F_SETFD IS NOT SAME AS F_SETFL ABOVE!
+  //these two sets of flags should be the same
+  fcntl(S->in_fd, F_SETFD, val);
+  fcntl(S->out_fd, F_SETFD, val);
 }
 
 
-
-
 /*Set timeout for select calls within STREAM*/
 void STREAMSetTimeout(STREAM *S, int val)
 {
@@ -109,8 +127,7 @@ void STREAMSetFlushType(STREAM *S, int Type, int StartPoint, int BlockSize)
 {
 S->Flags &= ~(FLUSH_ALWAYS | FLUSH_LINE | FLUSH_BLOCK | FLUSH_BUFFER);
 S->Flags |= Type;
-if (StartPoint==0) S->StartPoint=BlockSize;
-else S->StartPoint=StartPoint;
+S->StartPoint=StartPoint;
 S->BlockSize=BlockSize;
 }
 
@@ -118,24 +135,23 @@ S->BlockSize=BlockSize;
 /* the file pointer to that position */
 void STREAMResizeBuffer(STREAM *S, int size)
 {
-	int PageSize;
+	if (S->Flags & SF_MMAP) return;
 
-	PageSize=getpagesize();
-	if (S->Flags & SF_SECURE) S->BuffSize=(size / PageSize + 1) * PageSize;
-  else S->BuffSize=size;
-
-	S->InputBuff =(char *) realloc(S->InputBuff,S->BuffSize);
-	S->OutputBuff=(char *) realloc(S->OutputBuff,S->BuffSize);
+	if (S->Flags & SF_SECURE) 
+	{
+		if (! (S->Flags & SF_WRONLY)) SecureRealloc(&S->InputBuff, S->BuffSize, size, SMEM_SECURE);
+		if (! (S->Flags & SF_RDONLY)) SecureRealloc(&S->OutputBuff, S->BuffSize, size, SMEM_SECURE);
+	}
+  else 
+	{
+		if (! (S->Flags & SF_WRONLY)) S->InputBuff =(char *) realloc(S->InputBuff,size);
+		if (! (S->Flags & SF_RDONLY)) S->OutputBuff=(char *) realloc(S->OutputBuff,size);
+	}
+	S->BuffSize=size;
 
 	if (S->InStart > S->BuffSize) S->InStart=0;
 	if (S->InEnd > S->BuffSize) S->InEnd=0;
 	if (S->OutEnd > S->BuffSize) S->OutEnd=0;
-
-	if (S->Flags & SF_SECURE) 
-	{
-		mlock(S->InputBuff, S->BuffSize);
-		mlock(S->OutputBuff, S->BuffSize);
-	}
 }
 
 
@@ -254,15 +270,37 @@ return(FALSE);
 int STREAMInternalFinalWriteBytes(STREAM *S, const char *Data, int DataLen)
 {
 fd_set selectset;
-int result=0, count=0;
+int result=0, count=0, len;
 struct timeval tv;
 
 if (! S) return(STREAM_CLOSED);
 if (S->out_fd==-1) return(STREAM_CLOSED);
 
+
+//if we are flushing blocks, then pad out to the blocksize
+if (S->Flags & FLUSH_BLOCK) 
+{
+DataLen=S->BlockSize;
+if (DataLen > S->OutEnd) DataLen=S->OutEnd;
+
+/*
+ 	if (DataLen < S->BlockSize)
+	{
+	if (DataLen==0) DataLen=S->BlockSize;
+	else
+	{
+	len=(DataLen / S->BlockSize) * S->BlockSize;
+	if (S->OutEnd > len) len+=S->BlockSize;
+	memset(S->OutputBuff+S->OutEnd,0,len - S->OutEnd);
+	DataLen=len;
+	}
+	}
+*/
+}
+
 while (count < DataLen)
 {
-if (S->Flags & SF_SSL)
+if (S->State & SS_SSL)
 {
 #ifdef HAVE_LIBSSL
 result=SSL_write((SSL *) STREAMGetItem(S,"LIBUSEFUL-SSL-CTX"), Data + count, DataLen - count);
@@ -284,22 +322,24 @@ else
 	}
 
 	if (S->Flags & SF_WRLOCK) flock(S->out_fd,LOCK_EX);
-	if (S->BlockSize && (S->BlockSize < (DataLen-count))) result=S->BlockSize;
-	else result=DataLen-count;
+	result=DataLen-count;
+	//if (S->BlockSize && (S->BlockSize < (DataLen-count))) result=S->BlockSize;
   result=write(S->out_fd, Data + count, result);
 	if (S->Flags & SF_WRLOCK) flock(S->out_fd,LOCK_UN);
 
+	//yes, we neglect to do a sync. The idea here is to work opportunisitically, flushing out those pages
+	//that have been written. We do a sync and another fadvise in STREAMClose
+	#ifdef POSIX_FADV_DONTNEED
+	if (S->Flags & SF_NOCACHE) posix_fadvise(S->out_fd, 0,0,POSIX_FADV_DONTNEED);
+	#endif
+
 }
 
-  if (result < 1 && ((errno !=EINTR) && (errno !=EAGAIN)) ) break;
+
+  if (result < 1 && (errno !=EINTR) ) break;
   if (result < 0) result=0;
   count+=result;
   S->BytesWritten+=result;
-	if (S->Flags & FLUSH_BUFFER)
-	{
-		S->StartPoint=0;
-		break;
-	}
 }
 
 
@@ -382,7 +422,7 @@ if (
 //data. Even if S->InStart > 0 (meaning there are 'sent' bytes in the buffer)
 //we consider S->InStart to be 0 as regards sizeing the buffer, because those
 //sent bytes are still there.
-S->InputBuff=SetStrLen(S->InputBuff,len + S->InEnd );
+if ((len + S->InEnd) > S->BuffSize) STREAMResizeBuffer(S, len + S->InEnd);
 memcpy(S->InputBuff + S->InEnd, InBuff, len);
 S->InEnd+=len;
 }
@@ -455,6 +495,26 @@ return(Stream);
 }
 
 
+
+int STREAMOpenMMap(STREAM *S, int offset, int len, int Flags)
+{
+char *ptr;
+int MProt=PROT_READ;
+
+if (S->InputBuff) free(S->InputBuff);
+if (Flags & (SF_WRONLY | SF_RDWR)) MProt |= PROT_WRITE;
+ptr=(char *) mmap(0, len, MProt, MAP_SHARED, S->in_fd, offset);
+if (ptr==MAP_FAILED) return(FALSE);
+S->InEnd=len;
+S->InStart=0;
+S->Flags |= SF_MMAP;
+S->InputBuff=ptr;
+if (Flags & SF_SECURE) mlock(ptr, len);
+
+return(TRUE);
+}
+
+
 STREAM *STREAMOpenFile(const char *FilePath, int Flags)
 {
 int fd, Mode=0;
@@ -465,66 +525,78 @@ if (Flags & SF_WRONLY) Mode=O_WRONLY;
 else if (Flags & SF_RDONLY) Mode=O_RDONLY;
 else Mode=O_RDWR;
 
-if (Flags & SF_CREATE) Flags |= O_CREAT;
+if (Flags & SF_CREATE) Mode |= O_CREAT;
 
 if (strcmp(FilePath,"-")==0)
 {
 Stream=STREAMFromDualFD(0,1);
 Stream->Path=CopyStr(Stream->Path,FilePath);
-return(Stream);
+Stream->Flags=Flags;
 }
-
-fd=open(FilePath, Mode, 0600);
-if (fd==-1) return(NULL);
-
-if (Flags & SF_WRLOCK)
+else
 {
-	if (flock(fd,LOCK_EX | LOCK_NB)==-1)
+	fd=open(FilePath, Mode, 0600);
+	if (fd==-1) return(NULL);
+
+	if (Flags & SF_WRLOCK)
 	{
-		close(fd);
-		return(NULL);
+		if (flock(fd,LOCK_EX | LOCK_NB)==-1)
+		{
+			close(fd);
+			return(NULL);
+		}
+	}
+
+	if (Flags & SF_RDLOCK)
+	{
+		if (flock(fd,LOCK_SH | LOCK_NB)==-1)
+		{
+			close(fd);
+			return(NULL);
+		}
+	}
+
+
+	// check for symlink naughtyness. Basically a malicious user can
+	// try to guess the name of the file we are going to open in order
+	// to get us to write somewhere other than intended.
+	
+	
+	if (! (Flags & SF_SYMLINK_OK))
+	{
+		if (lstat(FilePath, &myStat) !=0)
+		{
+			close(fd);
+			return(NULL);
+		}
+	
+		if (S_ISLNK(myStat.st_mode))
+		{
+			syslog(LOG_USER | LOG_WARNING, "STREAMOpenFile Opened symlink when trying to open %s. Possible DOS attack?",FilePath);
+	 		close(fd);
+	 		return(NULL);
+		}
+	}
+	else 
+	{
+		stat(FilePath, &myStat);
+	}
+
+	//CREATE THE STREAM OBJECT !!
+	Stream=STREAMFromFD(fd);
+	STREAMSetTimeout(Stream,0);
+
+	if ( (Flags & (SF_RDONLY | SF_MMAP)) == (SF_RDONLY | SF_MMAP) ) STREAMOpenMMap(Stream, 0, myStat.st_size, Flags);
+	else
+	{
+	if (Flags & SF_TRUNC) ftruncate(fd,0);
+	if (Flags & SF_APPEND) lseek(fd,0,SEEK_END);
 	}
 
 }
-
-if (Flags & SF_RDLOCK)
-{
-	if (flock(fd,LOCK_SH | LOCK_NB)==-1)
-	{
-		close(fd);
-		return(NULL);
-	}
-}
-
-
-
-// check for symlink naughtyness. Basically a malicious user can
-// try to guess the name of the file we are going to open in order
-// to get us to write somewhere other than intended.
-
-
-if (! (Flags & SF_SYMLINK_OK))
-{
-	if (lstat(FilePath, &myStat) !=0)
-	{
-		close(fd);
-		return(NULL);
-	}
-
-	if (S_ISLNK(myStat.st_mode))
-	{
-		syslog(LOG_USER | LOG_WARNING, "STREAMOpenFile Opened symlink when trying to open %s. Possible DOS attack?",FilePath);
- 		close(fd);
- 		return(NULL);
-	}
-}
-
-if (Flags & SF_TRUNC) ftruncate(fd,0);
-if (Flags & SF_APPEND) lseek(fd,0,SEEK_END);
-
-Stream=STREAMFromFD(fd);
+Stream->Flags |= Flags;
+if (Stream->Flags & SF_SECURE) STREAMResizeBuffer(Stream, Stream->BuffSize);
 Stream->Path=CopyStr(Stream->Path,FilePath);
-STREAMSetTimeout(Stream,0);
 STREAMSetFlushType(Stream,FLUSH_FULL,0,0);
 
 return(Stream);
@@ -534,6 +606,7 @@ return(Stream);
 STREAM *STREAMClose(STREAM *S)
 {
 ListNode *Curr;
+int val;
 
 if (! S) return(NULL);
 
@@ -545,29 +618,53 @@ if (
 	 (strcmp(S->Path,"-") !=0)
 	)
 {
-if ((S->out_fd != -1) && (S->out_fd != S->in_fd)) close(S->out_fd);
-if (S->in_fd != -1) close(S->in_fd);
+	if ((S->out_fd != -1) && (S->out_fd != S->in_fd)) 
+	{
+		#ifdef POSIX_FADV_DONTNEED
+		if (S->Flags & SF_NOCACHE)
+		{
+	  		fdatasync(S->out_fd);
+ 	 		posix_fadvise(S->out_fd, 0,0,POSIX_FADV_DONTNEED);
+		}
+		#endif
+	
+		close(S->out_fd);
+	}
+
+	if (S->in_fd != -1)
+	{
+		#ifdef POSIX_FADV_DONTNEED
+		if (S->Flags & SF_NOCACHE) posix_fadvise(S->in_fd, 0,0,POSIX_FADV_DONTNEED);
+		#endif
+
+		close(S->in_fd);
+	}
 }
 
 Curr=ListGetNext(S->Values);
 while (Curr)
 {
-if (strncmp(Curr->Tag,"HelperPID",9)==0) kill(atoi(Curr->Item),SIGKILL);
-Curr=ListGetNext(Curr);
+	if (strncmp(Curr->Tag,"HelperPID",9)==0)
+	{
+		val=atoi(Curr->Item);
+ 		if (val > 1) kill(val, SIGKILL);
+	}
+	Curr=ListGetNext(Curr);
 }
 
 if (S->Flags & SF_SECURE) 
 {
-	munlock(S->InputBuff, S->BuffSize);
-	munlock(S->OutputBuff, S->BuffSize);
-	xmemset(S->InputBuff,0,S->BuffSize);
-	xmemset(S->OutputBuff,0,S->BuffSize);
+	SecureDestroy(S->InputBuff,S->BuffSize);
+	SecureDestroy(S->OutputBuff,S->BuffSize);
+}
+else
+{
+	if (! (S->Flags & SF_MMAP)) DestroyString(S->InputBuff);
+	DestroyString(S->OutputBuff);
 }
 
 ListDestroy(S->Values,(LIST_ITEM_DESTROY_FUNC)DestroyString);
 ListDestroy(S->ProcessingModules,DataProcessorDestroy);
-DestroyString(S->InputBuff);
-DestroyString(S->OutputBuff);
 DestroyString(S->Path);
 free(S);
 
@@ -575,17 +672,17 @@ return(NULL);
 }
 
 
-int STREAMDisassociateFromFD(STREAM *Stream)
+int STREAMDisassociateFromFD(STREAM *S)
 {
 int fd;
 
-if (! Stream) return(-1);
-fd=Stream->in_fd;
-STREAMFlush(Stream);
-DestroyString(Stream->InputBuff);
-DestroyString(Stream->OutputBuff);
-DestroyString(Stream->Path);
-free(Stream);
+if (! S) return(-1);
+fd=S->in_fd;
+STREAMFlush(S);
+if (! (S->Flags & SF_MMAP)) DestroyString(S->InputBuff);
+DestroyString(S->OutputBuff);
+DestroyString(S->Path);
+free(S);
 return(fd);
 }
 
@@ -604,8 +701,18 @@ void *SSL_CTX=NULL;
 
 if (! S) return(0);
 
+//we don't read from and embargoed stream. Embargoed is a state that we
+//use to indicate a stream must be ignored for a while
 if (S->State & SS_EMBARGOED) return(0);
 
+//we don't read to mmaped streams. We just update pointers to the mmap
+if (S->Flags & SF_MMAP)
+{
+	result=S->InEnd-S->InStart;
+	if (result < 1) return(STREAM_CLOSED);
+	return(result);
+}
+
 
 if (S->InStart >= S->InEnd)
 {
@@ -634,7 +741,7 @@ SSL_CTX=STREAMGetItem(S,"LIBUSEFUL-SSL-CTX");
 //if there are bytes available in the internal OpenSSL buffers, when we don't have to 
 //wait on a select, we can just go straight through to SSL_read
 #ifdef HAVE_LIBSSL
-if (S->Flags & SF_SSL)
+if (S->State & SS_SSL)
 {
 if (SSL_pending((SSL *) SSL_CTX) > 0) WaitForBytes=FALSE;
 }
@@ -681,7 +788,7 @@ if (read_result==0)
 	tmpBuff=SetStrLen(tmpBuff,S->BuffSize-S->InEnd);
 
 	#ifdef HAVE_LIBSSL
-	if (S->Flags & SF_SSL)
+	if (S->State & SS_SSL)
 	{
 		read_result=SSL_read((SSL *) SSL_CTX, tmpBuff, S->BuffSize-S->InEnd);
 	}
@@ -731,6 +838,7 @@ int bytes;
 	bytes=S->InEnd - S->InStart;
 	
 	if (bytes > DestSize) bytes=DestSize;
+	if (bytes < 1) return(0);
 	
 	memcpy(Dest,S->InputBuff+S->InStart,bytes);
 	S->InStart+=bytes;
@@ -769,7 +877,8 @@ while (total < Buffsize)
 		//we didn't have enough to satisfy another read like the one we just had
 	
 		//We must check for '< 1' rather than '-1' because 
-		result=FDCheckForBytes(S->in_fd);
+		if (S->Flags & SF_MMAP) result=-1;
+		else result=FDCheckForBytes(S->in_fd);
 
 		if (result ==-1) 
 		{
@@ -797,6 +906,7 @@ double STREAMTell(STREAM *S)
 {
 double pos;
 
+if (S->Flags & SF_MMAP) return((double) S->InStart);
 if (S->OutEnd > 0) STREAMFlush(S);
 
 #ifdef _LARGEFILE64_SOURCE
@@ -815,6 +925,17 @@ double STREAMSeek(STREAM *S, double offset, int whence)
 double pos;
 int wherefrom;
 
+if (S->Flags & SF_MMAP)
+{
+  switch (whence)
+  {
+  case SEEK_SET: S->InStart=offset; break;
+  case SEEK_CUR: S->InStart += offset; break;
+  case SEEK_END: S->InStart=S->InEnd-offset; break;
+  }
+  return((double) S->InStart);
+}
+
 if (S->OutEnd > 0) STREAMFlush(S);
 
 if (whence==SEEK_CUR) 
@@ -883,10 +1004,11 @@ return(AllDataWritten);
 }
 
 
-
+//this function returns the number of bytes *queued*, not number
+//written
 int STREAMInternalQueueBytes(STREAM *S, const char *Bytes, int Len)
 {
-int o_len, written=0, avail, val, result=0;
+int o_len, queued=0, avail, val=0, result=0;
 const char *ptr;
 
 	o_len=Len;
@@ -894,34 +1016,40 @@ const char *ptr;
 
 	do
 	{
-		avail=S->BuffSize - S->OutEnd;
-		if (avail > 0)
+		if (o_len > 0)
 		{
-		if (avail > o_len) avail=o_len;
+			avail=S->BuffSize - S->OutEnd;
+			if (avail > 0)
+			{
+			if (avail > o_len) avail=o_len;
 	
-		memcpy(S->OutputBuff+S->OutEnd,ptr,avail);
-		ptr+=avail;
-		o_len-=avail;
-		S->OutEnd+=avail;
+			memcpy(S->OutputBuff+S->OutEnd,ptr,avail);
+			ptr+=avail;
+			o_len-=avail;
+			S->OutEnd+=avail;
+
+			queued+=avail;
+			}
 		}
 	
-		if (
-					(S->OutEnd > S->StartPoint) ||
-					(S->OutEnd >= S->BuffSize) 
+		//Buffer Full, Write some bytes!!!
+		if ( (S->OutEnd > S->StartPoint) &&
+			 	(
+					 (S->OutEnd >= S->BuffSize) ||
+					 (S->Flags & FLUSH_ALWAYS) ||
+					 ((S->Flags & FLUSH_BLOCK) && (S->OutEnd > S->BlockSize))
+				)
 			)
 		{
-		//Buffer Full, Write some bytes!!!
-		if (S->BlockSize) val=(S->OutEnd / S->BlockSize) * S->BlockSize;
-		else val=S->OutEnd;
-	
-		if (val > 0) result=STREAMInternalFinalWriteBytes(S, S->OutputBuff, val);
-		if (result==0) written=STREAM_TIMEOUT;
+		//Any decision about how much to write takes place in InternalFinalWrite
+		result=STREAMInternalFinalWriteBytes(S, S->OutputBuff, S->OutEnd);
 		if (result < 1) break;
-		written+=result;	
+		S->StartPoint=0;	
 		}
 	} while (o_len > 0);
 
-return(written);
+if ((result < 0) && (result != STREAM_TIMEOUT)) return(result);
+return(queued);
 }
 
 
@@ -933,8 +1061,7 @@ int STREAMWriteBytes(STREAM *S, const char *Data, int DataLen)
 {
 const char *i_data;
 char *TempBuff=NULL;
-int len, written=0, result=0;
-int AllDataWritten=FALSE;
+int len, result=0;
 
 
 if (! S) return(STREAM_CLOSED);
@@ -945,58 +1072,21 @@ if (S->State & SS_WRITE_ERROR) return(STREAM_CLOSED);
 i_data=Data;
 len=DataLen;
 
-
-while (! AllDataWritten)
+if (ListSize(S->ProcessingModules))
 {
-  AllDataWritten=TRUE;
-
-	if (ListSize(S->ProcessingModules))
-	{
-		STREAMInternalPushProcessingModules(S, i_data, len, &TempBuff, &len);
-		i_data=TempBuff;
-	}
-
-	result=STREAMInternalQueueBytes(S, i_data, len);
-	if (result < 0) 
-	{
-		DestroyString(TempBuff);
-		return(result);
-	}
-	//Whatever happened above, o_data/o_len should hold data to be written
-	
-	
-	//Must do this to avoid sending data into the queue multiple times!
-	len=0;
-}
-
-
-//We always claim to have written the data that we've accepted
-//Though this can be overridden below
-written=DataLen;
-
-
-
-//if we are told to write zero bytes, that's a flush
-if ((S->OutEnd > 0) && ((DataLen==0) || (S->Flags & FLUSH_ALWAYS)))
-{
-	//if we are flushing blocks, then pad out to the blocksize
-	if (S->Flags & FLUSH_BLOCK) 
-	{
-		len=(S->OutEnd / S->BlockSize) * S->BlockSize;
-		if (S->OutEnd > len) len+=S->BlockSize;
-		memset(S->OutputBuff+S->OutEnd,0,len - S->OutEnd);
-		S->OutEnd=len;
-	}
-
-	result=STREAMInternalFinalWriteBytes(S, S->OutputBuff, S->OutEnd);
-	if (result < 0) written=result;
+	STREAMInternalPushProcessingModules(S, i_data, len, &TempBuff, &len);
+	i_data=TempBuff;
 }
 
+//This always queues all the data, though it may not flush it all
+//thus the calling application always believes all data is written
+//Thus we only report errors if len==0;
+if (len > 0) result=STREAMInternalQueueBytes(S, i_data, len);
+else if (S->OutEnd > S->StartPoint) STREAMInternalFinalWriteBytes(S, S->OutputBuff, S->OutEnd);
 
 DestroyString(TempBuff);
 
-
-return(written);
+return(result);
 }
 
 
@@ -1123,7 +1213,6 @@ while (1)
 		if (ptr) len=(ptr+1)-(S->InputBuff+S->InStart);
 		else len=S->InEnd-S->InStart;
 	
-		//Get length of RetStr, because SetStrLen might realloc it
 		RetStr=SetStrLen(RetStr,bytes_read + len);
 		len=STREAMTransferBytesOut(S, RetStr+bytes_read , len);
 		bytes_read+=len;
@@ -1137,6 +1226,7 @@ while (1)
 	if (S->InStart >= S->InEnd)
 	{
 		if (result==STREAM_TIMEOUT) return(RetStr);
+		if (result==STREAM_NODATA) return(RetStr);
 
 		if (bytes_read==0)
 		{
@@ -1258,78 +1348,233 @@ else ListAddNamedItem(S->Items,Name,Value);
 }
 
 
-int STREAMSendFile(STREAM *In, STREAM *Out, off_t Max)
+int STREAMFindCompare(const char *Line, const char *Item, const char *Delimiter, char **RetStr)
 {
-int bytes_read=0, result;
+char *ptr=NULL;
+int result;
+
+  if (StrValid(Delimiter)) ptr=strstr(Line, Delimiter);
+
+  if (ptr) result=strncmp(Item, Line, ptr-Line);
+  else result=strcmp(Item, Line);
+
+  if (RetStr && (result==0)) *RetStr=CopyStr(*RetStr, Line);
+
+return(result);
+}
+
+
+
+int STREAMFindBinarySearch(STREAM *S, const char *Item, const char *Delimiter, char **RetStr)
+{
+char *Tempstr=NULL;
+struct stat Stat;
+double Size, delta, pos;
+int result, inchar;
+
+
+if (S->Flags & SF_MMAP) Size=S->InEnd;
+else
+{
+	fstat(S->in_fd, &Stat);
+	Size=Stat.st_size;
+}
+
+delta=Size / 2;
+pos=delta;
+while (delta > 100)
+{
+	STREAMSeek(S, pos, SEEK_SET);
+	inchar=STREAMReadChar(S);
+	while ((inchar != '\n') && (inchar != EOF)) inchar=STREAMReadChar(S);
+	if (inchar==EOF) return(FALSE);
+	Tempstr=STREAMReadLine(Tempstr, S);
+	StripTrailingWhitespace(Tempstr);
+	result=STREAMFindCompare(Tempstr, Item, Delimiter, RetStr);
+	if (result ==0) break;
+
+	delta /= 2;
+	if (result > 0) pos+=delta;
+	else pos-=delta;
+}
+
+//rewind if we need to
+while ((result < 0) && (pos > 0))
+{
+pos-=1024;
+//if we're near the start of the file, then just seek back to the start
+if (pos < 1024) pos=0;
+STREAMSeek(S, pos, SEEK_SET);
+if (pos > 0)
+{
+inchar=STREAMReadChar(S);
+while ((inchar != '\n') && (inchar != EOF)) inchar=STREAMReadChar(S);
+}
+
+Tempstr=STREAMReadLine(Tempstr, S);
+StripTrailingWhitespace(Tempstr);
+
+result=STREAMFindCompare(Tempstr, Item, Delimiter, RetStr); 
+}
+
+while (Tempstr)
+{
+	result=STREAMFindCompare(Tempstr, Item, Delimiter, RetStr); 
+	if (result < 1)
+	{
+	DestroyString(Tempstr);
+	if (result==0) return(TRUE);
+	return(FALSE);
+	}
+	Tempstr=STREAMReadLine(Tempstr, S);
+	StripTrailingWhitespace(Tempstr);
+}
+
+DestroyString(Tempstr);
+return(FALSE);
+}
+
+
+int STREAMFindItem(STREAM *S, const char *Item, const char *Delimiter, char **RetStr)
+{
+char *Tempstr=NULL;
+
+if (!S) return(FALSE);
+
+if (S->Flags & SF_SORTED) return(STREAMFindBinarySearch(S, Item, Delimiter, RetStr));
+
+Tempstr=STREAMReadLine(Tempstr, S);
+while (Tempstr)
+{
+	StripTrailingWhitespace(Tempstr);
+	if (STREAMFindCompare(Tempstr, Item, Delimiter, RetStr)==0)
+	{
+		DestroyString(Tempstr);
+		return(TRUE);
+	}
+	Tempstr=STREAMReadLine(Tempstr, S);
+}
+
+DestroyString(Tempstr);
+return(FALSE);
+}
+
+
+
+
+unsigned long STREAMSendFile(STREAM *In, STREAM *Out, unsigned long Max, int Flags)
+{
+unsigned long bytes_transferred=0;
+int result;
 int UseSendFile=FALSE;
 
 //val has to be as long as possible, because it will hold the difference
-//between two off_t values. However, use of 'long long' resulted in an
+//between two long values. However, use of 'long long' resulted in an
 //unsigned value, which caused all manner of problems, so a long is the
 //best we can manage
-long val, len; 
+long val, len;
+
+
+
+//if we are not using ssl and not using processor modules, we can use
+//kernel-level copy!
+
+if (Max==0) len=BUFSIZ;
+else len=(long) (Max-bytes_transferred);
 
 
-#undef USE_SENDFILE
 #ifdef USE_SENDFILE
 #include <sys/sendfile.h>
 
-//if we are not using ssl and not using processor modules, we can use 
-//kernel-level copy!
 if (
-			(! (In->Flags & SF_SSL)) && 
-			(! (Out->Flags & SF_SSL)) && 
-			(ListSize(In->ProcessingModules)==0) && 
-			(ListSize(Out->ProcessingModules)==0)
-	) UseSendFile=TRUE;
+      (Flags & SENDFILE_KERNEL) &&
+      (! (In->State & SS_SSL)) &&
+      (! (Out->State & SS_SSL)) &&
+      (ListSize(In->ProcessingModules)==0) &&
+      (ListSize(Out->ProcessingModules)==0)
+)
+{
+  UseSendFile=TRUE;
+  STREAMFlush(Out);
+}
 #endif
 
-len=(long) Max-bytes_read;
 
+while (len > 0)
+{
 #ifdef USE_SENDFILE
 	if (UseSendFile)
-	{
-		STREAMFlush(Out);
-		result=sendfile(Out->out_fd, In->in_fd,0,len);
-		if ((result < 1) && (bytes_read==0)) UseSendFile=FALSE;
-	}
+  {
+    result=sendfile(Out->out_fd, In->in_fd,0,len);
+    if (result < 1)
+    {
+      if (bytes_transferred==0) UseSendFile=FALSE;
+      result=STREAM_CLOSED;
+    }
+  	else 
+		{
+			//if used sendfile we need to do this because we didn't call
+			//lower level functions like 'STREAMReadCharsToBuffer' that would
+			//update these
+			In->BytesRead+=result;
+			Out->BytesWritten+=result;
+		}
+  }
 #endif
 
-	if (! UseSendFile)
+  if (! UseSendFile)
+  {
+    //How much still left to transfer?
+
+    //if outbuff is full do some flushing
+    val=Out->BuffSize - Out->OutEnd;
+    if (val < 1)
+    {
+      STREAMFlush(Out);
+      sleep(0);
+      break;
+      val=Out->BuffSize - Out->OutEnd;
+    }
+
+    result=0;
+
+		
+	val=In->InEnd - In->InStart;
+	//only read if we don't already have some in inbuff
+	if (val < 1)
 	{
-		//How much still left to transfer?
-
-		//if outbuff is full do some flushing
-		val=Out->BuffSize - Out->OutEnd;
-		if (val < 1)
-		{
-			STREAMFlush(Out);
-			sleep(0);
-			val=Out->BuffSize - Out->OutEnd;
-		}
-
-		//if outbuff smaller than len, then shrink len
-		if (len > val) len=val;
-
-		//if inbuff remaining is smaller than len, then shrik that
-  	STREAMReadCharsToBuffer(In);
+		result=STREAMReadCharsToBuffer(In);
 		val=In->InEnd - In->InStart;
-		if (len > val) len=val;
-		if (len <0) len=0;
+	}
 
-		result=STREAMWriteBytes(Out,In->InputBuff+In->InStart,len);
+	if (len > val) len=val;
+
+	//nothing to write!
+	if (len < 1)
+	{
+		//nothing in either buffer! Stream empty. Is it closed?
+		if ((Out->OutEnd==0) && (result==STREAM_CLOSED)) break;
+		len=0;
+	}
+
+	result=STREAMWriteBytes(Out,In->InputBuff+In->InStart,len);
+
+	if (result > 0)
+	{
+  In->InStart+=result;
+  bytes_transferred+=result;
+	}
 
-		if (result > 0)
-		{
-  	In->InStart+=result;
-		bytes_read+=result;
-		}
-		else if (result ==STREAM_CLOSED) 
-		{
-			if (bytes_read==0) bytes_read=result;
-		}
 	}
 
 
-return(bytes_read);
+if (! (Flags & SENDFILE_LOOP)) break;
+
+if (Max==0) len=BUFSIZ;
+else len=(long) (Max-bytes_transferred);
 }
+
+if ((result==STREAM_CLOSED) && (bytes_transferred==0) ) return(STREAM_CLOSED);
+return(bytes_transferred);
+}
+
